
(* Signatur fuer die Datenbank *)

signature Database = sig
 
  type ''a Tree
  type ''a DBLTree
  type ''a Database
  eqtype 'a Ref

  type 'a IList
  type 'a ListNode

  type id = int
  datatype ''a dbsig = 
      L of ''a               (* Standardknoten *) 
    | D of ''a * id          (* Interner Datenbankknoten *)
    | P of id                (* New Strongly Connected Compoment *)

  (* Erstellt eine leere Datenbank *)
  val createDatabase : (''a * ''a -> order) -> (''a -> int) -> ''a Database

  (* Liefert die Anzahl der ID's zurck *)
  val dbgetids : ''a Database -> int

  val cmpdbsig : (''a * ''a -> order) -> (''a dbsig * ''a dbsig -> order)

  (* Fgt einen Baum zur Datenbank hinzu und liefert die ID zurck. *)
  val dbinsert : ''a Database -> ''a dbsig DBLTree Ref -> int

  (* Gibt die ID eines Baumes zurck wenn er in der Datenbank
     ist, andernfalls NONE *)
  val dbgetid : ''a dbsig DBLTree Ref -> ''a Database -> int option

  (* Liest den zu einer ID gehrigen Baum aus der Datenbank aus
     Es sind keine P-Knoten mehr enthalten *)
  val dbgettreefromid : int -> ''a Database -> ''a dbsig Tree Ref option
  
  (* Liest den Baum aus der Datenbank aus. P-Knoten bleiben erhalten *)
  val dbgetdbtreefromid : int -> ''a Database -> ''a dbsig DBLTree Ref option

  (* Setzt und liest den Namen eines Baumes anhand der ID *)
  val dbsetname : int -> string -> ''a Database -> unit
  val dbgetname : int -> ''a Database -> string option

  (* Liefert die ID eines benannten Baumes zurck *)
  val dbgetidfromname : string -> ''a Database -> int option

end where type ''a DBLTree = ''a Tree.DBLTree
      and type ''a Tree = ''a Tree.Tree
      and type 'a Ref = 'a Ref.Ref
      and type 'a IList = 'a IList.IList
      and type 'a ListNode = 'a IList.ListNode
