(*********************************************************
*
* Imlementierung eines Hashingbaumes
* Es werden Listen ber einem beliebigen Datentyp
* untersttzt.
*
**********************************************************)

structure Hash :> Hash = struct

open NaiveMap;

(* Datentyp fuer den Hash Baum. Jeder Knoten besteht aus
   den Daten vom Typ ''b und einer Funktion von ''a nach
   (''a,''b) hash die die von dem Knoten ausgehenden Kanten
   darstellen *)

datatype (''a,''b,''c) hash' = NN of (''b option  * (''c, (''a,''b,''c) hash' ref) map ref * (''a -> ''c list))
type (''a,''b, ''c) hash = ((''a, ''b, ''c) hash') ref

(* Erstellt einen leeren Hash Baum *)

fun mkTable f = ref(NN(NONE,ref emptymap,f)) 

(* Folgt einer Kante im Hash Baum falls vorhanden *)

fun follow p h = case !h of
       (NN(id,m,f)) => lookup (!m,p)

(* Folgt einer Kante im Hash Baum falls vorhanden,
   andernfalls wird eine neue Kante erstellt. *)

fun node_insert p h = case follow p h of
    (SOME h') => h'
   | NONE => (case !h of
       (NN(id,m,f)) => let val t = ref (NN(NONE, ref emptymap,f))
                     in (m:=insertmap(!m,(p,t)); t) end)

(* Fuegt einen Datensatz vom Typ ''b ein, und zwar
   an die Stelle im Baum, zu der der Pfade vom Typ
   ''a list fuehrt. *)

fun peekInsert' (h,nil,data) = (case !h of (NN(NONE,m,f)) => (h:=NN(SOME data,m,f); NONE)
                              |(NN(SOME data',m,f)) => SOME data')
  | peekInsert' (h,p::pr,data) = let val h' = node_insert p h
                                in peekInsert' (h',pr,data) end

fun peekInsert h (t,data) = 
    case !h of (NN(_,_,f)) =>
	peekInsert'(h,f t,data)

(* Liest vorher geschriebene Daten wieder aus dem Hash Baum *)

fun peek' (h,nil) = (case !h of (NN(NONE,m,f)) => NONE
                        |(NN(SOME data',m,f)) => SOME data')
  | peek' (h,p::pr) = let val x = follow p h
		     in case x of NONE => NONE
		        |(SOME h') => peek' (h',pr) 
		     end

fun peek h t = 
    case !h of (NN(_,_,f)) =>
	peek'(h,f t)
end
