(* Imperative Liste *)

structure IList :> IList = 
struct

  datatype 'a node' = E
                   | FIRST of 'a node' ref
                   | N of 'a node' ref * 'a * 'a node' ref
                   | LAST of 'a node' ref
  type 'a node = 'a node' ref
  type 'a ListNode = 'a node

  (* speichert ersten und letzten Knoten sowie die Lnge der Liste*)
  type 'a IList = ('a node * 'a node * int ref) ref

  fun empty _ = let val fst = ref(E)
                    val lst = ref(E)
                    val _ = fst := FIRST lst
                    val _ = lst := LAST fst
                in ref (fst,lst,ref 0) end
  

  fun decCounter l = case !l of (_,_,c) => c := !c - 1
  fun incCounter l = case !l of (_,_,c) => c := !c + 1

  fun size l = case !l of (_,_,n) => !n

  fun isEmpty l = size l = 0

  fun next n = case !n of (N(_,_,nxt)) => nxt 
               | (FIRST nxt) => nxt
               | _ => raise Empty

  fun prev n = case !n of (N(prv,_,_)) => prv 
               | (LAST prv) => prv
               | _ => raise Empty 

  fun first l = case !l of (f,_,_) => next f
  fun last l = case !l of (_,e,_) => e

  fun value n = case !n of (N(_,v,_)) => v 
               | _ => raise Empty

  fun isFirst n = case !n of (FIRST _) => true | _ => false
  fun isEnd n = case !n of (LAST _) => true | _ => false

  (* Verbindet Knoten n0 mit n1, z.B. um nach dem Lschen eines Knotens die Liste zusammenzuhalten *)
  fun link n0 n1 = (case !n0 of (N(prv,x,nxt)) => n0 := N(prv,x,n1)
                    | (FIRST nxt) => n0 := FIRST(n1)
                    | _ => ();
                    case !n1 of (N(prv,x,nxt)) => n1 := N(n0,x,nxt)
                    | (LAST prv) => n1 := LAST(n0)
                    | _ => ())

  fun clear l = l:=(!(empty()))

  fun erase (l,n) = 
      case !n of 
	  (N(prv,x,nxt)) => (decCounter(l); link prv nxt; x)
	| _ => raise Empty

  fun insert (l,n,x) = 
    let val newnode = ref (N(n,x,n))
    in case !n of (N(prv,_,_)) => (link prv newnode;
                                   link newnode n; incCounter(l))
       | (FIRST nxt) => (link n newnode;
                         link newnode nxt; incCounter(l))
       | (LAST prv) => (link prv newnode;
                        link newnode n; incCounter(l))
       | E => raise Empty;
       newnode
    end
  
  fun front l = value(first(l))
  fun back l = value(prev(last(l)))
  fun pushFront (l,x) = insert (l,first(l),x) 
  fun pushBack (l,x) = insert (l,last(l),x) 
  fun popFront l = erase (l,first(l))
  fun popBack l = erase (l,prev(last(l)))

  fun imap f l = 
    let val node = ref (first(l))
    in while not (isEnd(!node)) do
	(f (!node);
        node := next (!node))
    end

  fun copy l =
      let val node = ref (first(l))
          val new = empty()
      in while not (isEnd(!node)) do
	  (pushBack(new, value(!node));
	   node := next (!node));
	 new
      end

end
