structure Interface :> Interface = struct

type tree = int

open Parser
open Database
open TypeTrees
open Splaymap
open Ref
open Tree

(* Datentyp fr Typen *)

datatype exp =
    Const of string
  | App of exp*exp
  | Tree of tree
  | Var of string
  | Mu of string * exp
  | Lambda of string * exp

(* ber dieser Signatur werden die zyklischen
   Typenbume erstellt *)

datatype typ =
    TC of string
  | TA
  | BACK of int
  | TV of string
  | TL

(* Hilfsdatentyp *)

datatype binder =
    LAMBDA of int
  | MU of typ dbsig Tree Ref

(* Ausnahmen *)

exception NotClosed
exception NotInDB
exception InternalDBError

fun Compare ((TC s),(TC s')) = String.compare(s,s')
  | Compare ((TC _),_      ) = LESS
  | Compare (_     ,(TC _) ) = GREATER
  | Compare (TA    ,TA     ) = EQUAL
  | Compare (TA    ,_      ) = LESS
  | Compare (_     ,TA     ) = GREATER
  | Compare ((BACK i),(BACK j)) = Int.compare(i,j)
  | Compare ((BACK _),_       ) = LESS
  | Compare (_       ,(BACK _)) = GREATER
  | Compare ((TV s),(TV s')) = String.compare(s,s')
  | Compare ((TV _),_      ) = LESS
  | Compare (_     ,(TV _) ) = GREATER
  | Compare (TL,TL) = EQUAL
   
fun StringtoInt s = List.foldl (fn (d,num) => (num*10+Char.ord(d)) mod 1234567) 0 (explode s)

fun ToInt (TC string) = StringtoInt string
  | ToInt (TV string) = ~(StringtoInt string)
  | ToInt (TL) = 676763
  | ToInt (BACK num)  = ~(676763+num)
  | ToInt TA = 0

val db = createDatabase Compare ToInt;

(* Umwandlung Typ -> zyklischer Typenbaum *)

fun tytree' binder lcount (Const const):typ dbsig Tree Ref = Ref (N(L(TC const),nil))
  | tytree' binder lcount (App(e1,e2)) = Ref (N(L(TA),[tytree' binder lcount e1,tytree' binder lcount e2]))
  | tytree' binder lcount (Mu(s,e)) = let val node = Ref E: typ dbsig Tree Ref
				      in  node := !(tytree' (insert (binder,s,MU node)) lcount e);
					  case !node of E => Ref (N(L(TC "__BLACK_HOLE__"),nil))
					          (* raise InvalidType *)
					  | _ => node
				      end 
  | tytree' binder lcount (Lambda(s,e)) =
       Ref(N(L TL,[tytree' (insert (binder,s,LAMBDA lcount)) (lcount+1) e]))
  | tytree' binder lcount (Tree id) = Ref(N(P id, nil))
  | tytree' binder lcount (Var s) =
       (case Int.fromString s of (* Falls eine Variable eine Zahl darstellt
				    so wird der zugehrige Knoten der
				    Datenbank genommen !!!!!!!!!!!!!! *)
	    (SOME id) => Ref(N(P id,nil))
	  |  NONE => 
		case peek(binder, s) of
		    NONE => raise NotClosed (*Ref(N(TV s,nil))*)
		  |(SOME(MU n)) => n
		  |(SOME(LAMBDA i)) => Ref(N(L(BACK (lcount-i-1)),nil)))

(* Fgt einen Typ zur Datenbank hinzu und gibt den entsprechenden tree zurck *)

fun enter exp = dbinsert db (toDBLTree(tytree' (mkDict String.compare) 0 exp))

(* Gleicheit auf trees *)

val equal = op=

fun buildMu exp done t= case peek(!done,t) of
                             (SOME (SOME id)) => Mu("t"^Int.toString id,exp)
			   |  _               => exp

(* Packt genau EINE Partition aus der Datenbank wieder aus *)

fun return' done lambda mu (t:typ dbsig Tree Ref) =
	   case peek(!done, t) of
	       NONE =>
		   (done := insert(!done,t,NONE);
		    case !t of
			E => let val label = "t"^Int.toString (!mu)
			     in  mu := !mu + 1;
				 Mu(label, Var label)
			     end
		      |(N(D(TC x,_),_)) => if String.compare(x,"__BLACK_HOLE__") = EQUAL then
			                       let val label = "t"^Int.toString (!mu)
					       in  mu := !mu + 1;
						   Mu(label, Var label)
					       end else Const x
		      |(N(D(TV x,_),_)) => Var x
		      |(N(D(TL,_), [t']))  => 
			     let val node = Lambda("x"^(Int.toString lambda),return' done (lambda+1) mu t')
			     in  buildMu node done t
			     end
		      |(N(D(TA,_),[t',t''])) => 
			     let val node = App(return' done lambda mu t',return' done lambda mu t'')
			     in  buildMu node done t
			     end
		      |(N(D(BACK id,_),_)) => Var("x"^Int.toString(lambda-id)) 
	              |(N(P id,_)) => Tree id
		      | _ => raise InternalDBError  (* kann nicht auftreten *)
                   )
	     |(SOME NONE) =>
		   (done := insert(!done,t,SOME (!mu));
		    Var("t"^Int.toString (!mu)) before mu := !mu + 1)
	     |(SOME(SOME id)) => Var("t"^Int.toString id)

fun return id =
	   case dbgetdbtreefromid id db of 
	       NONE => raise NotInDB  (* kann nicht auftreten *)
	     |(SOME t) => return' (Ref (mkDict Ref.cmp)) 0 (Ref 0) (fromDBLTree t)

fun TypToString (App(App(Const "+",e),f)) = TypToString e ^ " + " ^ TypToString f
  | TypToString (Const x) = "Const(" ^ x ^")" 
  | TypToString (App(e,f)) = "(" ^ TypToString e ^ " " ^ TypToString f ^ ")"
  | TypToString (Tree id) = "Tree(#" ^ Int.toString id ^")"
  | TypToString (Var x) = "Var(" ^ x ^ ")"
  | TypToString (Mu (x,e)) = "(u " ^ x ^ "." ^ TypToString e ^ ")"
  | TypToString (Lambda (x,e)) = "(l " ^ x ^ "." ^ TypToString e ^ ")"
fun printtype e = print (TypToString e ^ "\n")

fun plus(a,b) = App(App(Const "+",a),b)
val nat = Mu("x",plus(Const "unit",Var "x"))

(*val _ = print (Int.toString(enter(Const "unit")) ^"\n");
val _ = print (Int.toString(enter(Const "unit")) ^"\n");
val _ = print (Int.toString(enter(Const "+")) ^"\n");
val _ = print (Int.toString(enter(nat)) ^"\n");
*)
(*val _ = print ("equal? "^ Bool.toString(equal (enter(nat), enter(plus(Const "unit",nat))))^"\n")
val _ = printtype (plus(Const "unit",nat))
val _ = printtype (return(enter(plus(Const "unit",Const "nat"))))
val _ = printtype (nat)
val _ = printtype (return(enter(nat)))
*)
end
