(*   Lexer   *)

structure Lex :> Lex = struct

exception Error of string

type id = string

datatype token =
    ADD          (* + *)
  | MUL          (* * *)
  | ARROW        (* -> *)
  | LPAR         (* ( *)
  | RPAR         (* ) *)
  | ID of string (* Variablen *)
  | UNIT         (* unit *)
  | NULL         (* null *)
  | MY           (* u t. M (rekursive Typen) *)
  | LAMBDA       (* l x. M (Lambda Terme)    *)
  | DOT          (* . *)
  | ASSIGN       (* = *)

(* Einlesen des Strings bis zum nchsten Trennzeichen *)
fun getString nil s = (nil,s)
  | getString (c::cr) s = if Char.isAlpha c orelse Char.isDigit c 
                          then getString cr (s ^ str c)
                          else (c::cr,s)

fun lex' s nil = rev s
  | lex' s (#" " :: cs)        = lex' s cs
  | lex' s (#"\n" :: cs)       = lex' s cs
  | lex' s (#"\t" :: cs)       = lex' s cs 
  | lex' s (#"+" :: cs)        = lex' (ADD::s) cs
  | lex' s (#"*" :: cs)        = lex' (MUL::s) cs
  | lex' s (#"-":: #">" :: cs) = lex' (ARROW::s) cs
  | lex' s (#"(" :: cs)        = lex' (LPAR::s) cs
  | lex' s (#")" :: cs)        = lex' (RPAR::s) cs
  | lex' s (#"." :: cs)        = lex' (DOT::s) cs
  | lex' s (#"=" :: cs)        = lex' (ASSIGN::s) cs
  | lex' s (c::cs) = 
      if Char.isAlpha c orelse Char.isDigit c then lexAlpha s (c::cs)
      else raise Error ("unknown character: " ^ str c)

and lexAlpha s cs =
    case getString cs "" of
	(cr,"unit") => lex' (UNIT::s)   cr
      | (cr,"null") => lex' (NULL::s)   cr
      | (cr,"u")    => lex' (MY::s)     cr
      | (cr,"l")    => lex' (LAMBDA::s) cr
      | (cr,st    ) => lex' (ID(st)::s) cr 

fun lex s = lex' nil (explode s)

end
