(********************************************************
*
* Benutzeroberflche der Typendatenbank
*
*********************************************************)

structure main :> main = struct

open Ref;
open TypeTrees;
open Lex;
open Parser;
open Tree;
open Database;
open TextIO;
open Timer;
open Dynlib;
open TextIO;
open Random;
open Interface;

datatype t_mode = STANDARD | INTERFACE
val mode = Ref STANDARD

exception Error of string

(* Dynamisch eingebundene Bibliothek zur Zeitmessung                        *)
(* Ntig aufgrund von Problemen mit den ML-internen Routinen                *)
(**)val libtimeit = dlopen{lib="./libtimeit.so", flag=RTLD_NOW, global=true}
(**)val c_timeit = dlsym libtimeit "timeit__Fv"
(**)fun gettime () : int = app1 c_timeit ()
(**)
(**)fun timeit f x = 
(**)    let val t0 = gettime()
(**)    in (f x, real(gettime() - t0)/1000.0)
(**)    end
	

(* Leere Datenbank erstellen *)
val db = createDatabase labelcmp label2int

(* Prueft, ob der Benutzer dem Baum einen Namen gegeben hat. *)
fun parse_name ((ID s)::ASSIGN::tl) = (tl,SOME s)
  | parse_name tl = (tl,NONE)

fun is1 s0 [(ID x0)] = x0=s0
  | is1 _ _ = false

fun is2 s0 s1 [(ID x0),(ID x1)] = x0=s0 andalso x1=s1
  | is2 _ _ _ = false

fun printTreeSize t = 
    let val (nodes,edges,maxedges) = getTreeSize t
    in print ("Knoten    : " ^ Int.toString nodes ^ "\n" ^
	      "Kanten    : " ^ Int.toString edges ^ "\n" ^
	      "Max Kanten: " ^ Int.toString maxedges ^ "\n\n")
    end

(* Testet die Reprsentation                                      *)
(*                                                                *)
(* Dazu werden Zufallsbume erzeugt, die dann einerseits          *)
(* - normal in die Datenbank eingefgt werden                     *)
(* und andererseits                                               *)
(* - zuerst minimiert und dann erst eingefgt werde.              *)
(* Beide Male muss der Baum dieselbe ID der Datenbank bekommen,   *)
(* wenn alles geklappt hat.                                       *)
fun verify_rep () =
    let
	val continue = Ref true
	val db = createDatabase Int.compare (fn x => x)
        val gen = newgen()
        val k = Ref 0
    in
	while !continue do
	    
	    let val t = randomTree (fn x => (k := !k + 1; 
                                             if !k mod 10 = 0 andalso dbgetids db > 0 
						 then P (trunc(random gen*real(dbgetids db))) 
                                                 else L x)) 
                          (fn x => case x of (P _) => true | _ => false)
                          (*(trunc(random gen*200.0))*) 30 4 2 

		(*val _ = printTreeSize t;
		print ("unshared tree: " ^ drawTree t (fn (L x) => Int.toString x | (R x)=> Int.toString x) ^ "\n")*)
		val t0 = copyTree t
		val t1 = copyTree t

		val s0 = dbinsert db (toDBLTree t0)
	        val st1 = share (cmpdbsig Int.compare) t1
                (*val _ = (print "shared size: \n"; printTreeSize st1)*)
		val s1 = dbinsert db (toDBLTree st1)

	    in
		if s0 <> s1 orelse s0 < 0 
                   orelse s1 < 0 (*orelse s2 < 0
                   orelse not(compareTrees(st1,valOf(dbgettreefromid s1 db))*) then
		    (print "\nError:\n";
		     print ("unshared tree: " ^ drawTree t0 (fn (L x) => "L"^Int.toString x 
		                                            | (P x)=> "P"^Int.toString x
			                                    | (D(x,id)) => "D"^Int.toString x) ^ "\n");
                     print ("shared tree: " ^ drawTree st1 (fn (L x) => "L"^Int.toString x 
		                                            | (P x)=> "P"^Int.toString x
			                                    | (D(x,id)) => "D"^Int.toString x) ^ "\n");
		     print ("shared tree: " ^ Int.toString(s0) ^ "\n");
		     print ("shared spec: " ^ Int.toString(s1) ^ "\n");
		     continue := false)
                else print "+"
	    end
    end

(* Testet den Share Algorithmus                                   *)
(*                                                                *)
(* Dazu werden Zufallsbume erzeugt, die dann einerseits          *)
(* - mit dem implementierten Share-Algorithmus                    *)
(* und andererseits                                               *)
(* - mit einem naiven Spezifikationsalgorithmus  minimiert werden *)
(* Beide Male muss derselbe (syntaktisch!) Baum als Ergebnis      *)
(* entstehen.                                                     *)
fun verify_share () =
    let
	val continue = Ref true
    in
	while !continue do
	    
	    let val t = randomTree (fn x => x) (fn x => false) 60 4 2

		val t0 = copyTree t
		val t1 = copyTree t    
		val s0 = share Int.compare t0
		val s1 = s0(*share_spec t1*)
	    in
		if compareTrees (s0,s1) then 
		    if compareTrees (s0,t) then print "-" else print "+"
		else
		    (print "\nError:\n";
		     print ("unshared tree: " ^ drawTree t (Int.toString) ^ "\n");
		     print ("shared tree: " ^ drawTree s0 (Int.toString) ^ "\n");
		     print ("shared spec: " ^ drawTree s1 (Int.toString) ^ "\n");
		     continue := false)
	    end
    end


(* Geschwindigkeitstest von share und share_spec                  *)
(*                                                                *)
(* Dazu werden Zufallsbume erzeugt, die dann sowohl mit dem      *)
(* share-Algorithmus als auch mit seiner Spezifikation minimiert  *)
(* werden. Die dazr bentigte Zeit wird ausgegeben.              *)
fun speed_test() =
    let
	val t = randomTree (fn x => x) (fn x => false) 200 4 2
        val _ = printTreeSize t
	val _ = print "\n"
	val t0 = copyTree t
	val t1 = copyTree t
    in
	let
	    val _ = print ("starte share\n")
	    val (s0,time0) = timeit (share Int.compare) t0
	    val _ = print ("Time : " ^ Real.toString time0 ^ "\n")
	    val _ = printTreeSize s0
 
	    val _ = print ("\nstarte share_spec\n")
       	    val (s1,time1) = timeit share_spec t1
	    val _ = print ("Time : " ^ Real.toString time1 ^ "\n")
            val _ = printTreeSize s1
	in
	    ()
	end
    end
	 
(* Geschwindigkeitstest von share und share_spec                   *)
(*                                                                 *)
(* Dazu werden Zufallsbume erzeugt, die dann sowohl mit dem       *)
(* share-Algorithmus als auch mit seiner Spezifikation minimiert   *)
(* werden. Die dazr bentigte Zeit wird in eine Datei geschrieben.*)
fun speed_save(fname,n0,n1,count) =
    let 
	val out = openAppend fname
	val db = createDatabase Int.compare (fn x => x)
	val i = Ref 0
	val gen = newgen()
	val time = Ref 0.0
    in 
	while !i < count do
	let
	    val t = randomTree (fn x => x) (fn x => false) (n0+trunc(random gen*real(n1-n0))) 4 2
	    val (nodes,edges,maxedges) = getTreeSize t
	    val _ = output (out,Int.toString nodes ^ ";" ^
				Int.toString edges ^ ";" ^
				Int.toString maxedges ^ ";")
	    val t0 = copyTree t
	    val t1 = copyTree t
  
	    (*val (s0,time0) = timeit (share Int.compare) t0*)
	    val (s1,time1) = timeit share_spec t1
	    (*val _ = time := !time + time0 + time1*)

            (*val (s0,time0) = timeit (dbinsert db) (toDBLTree t0)*)

	    (*val (nodes,edges,maxedges) = getTreeSize s0
	    val _ = output (out,Int.toString nodes ^ ";")*)
	    val _ = output (out,Real.toString time1 ^ "\n")
	    (*val _ = output (out,Real.toString time1 ^ "\n")*) 

	    val _ = flushOut out
	in 
	    i := !i +1;
	    print "*"
	end;
	closeOut out;
	print ("\nGesamt: " ^ Real.toString (!time) ^ "\n")
    end
	

(* Hier findet die Umrechnung von dem String s in einen         *)
(* Typ, zyklischen Graphen, dann das Einfuegen in die Datenbank *)
(* und die Ausgabe des Minimaltyps statt.                       *)

fun exec_standard typ name =
let
    val tree    = tytree db typ            (* Den Typ in einen zyklischen Graphen umwandeln        *)
    val _       = print ("Tree = " ^ drawTree (fromDBLTree tree) label2string ^ "\n")
    (*val _       = print ("Shared = " ^ drawTree (share tree) label2string ^ "\n")*)
    val id      = dbinsert db tree         (* den zyklischen Graphen in die Datenbank einfuegen    *)
    val _       = case name of (SOME n) => dbsetname id n db
                              | NONE    => ()
    val dbtree' = valOf(dbgetdbtreefromid id db)
    val _       = print ("db = " ^ drawTree (fromDBLTree dbtree') label2string ^ "\n")

    val tree' = valOf(dbgettreefromid id db)
in
    print ("id = " ^ (Int.toString id)  ^ "\n"); 
    (case dbgetname id db of
	 (SOME s) => print ("name = " ^ s ^ "\n")
       |  NONE    => ());
    print ("type = " ^ tytree2tystring db tree' ^ "\n");
    SOME id
end

fun exec_iface typ =
let
    fun cas(l,t1,t2) = App(App(Const l,t1),t2)
    fun cas_list' (lst,nil) = lst
      | cas_list' (lst,x::xr) = cas_list'(App(lst,x),xr)
    fun cas_list (l,xr) = cas_list'(Const l,xr)
  
    fun to_cascard Null = Const "Null"
      | to_cascard Unit = Const "Unit"
      | to_cascard (Arrow(t1,t2)) = cas("->",to_cascard t1, to_cascard t2)
      | to_cascard (Parser.App(t1,t2)) = raise Error "application not supported"
      | to_cascard (Add(t1,t2)) = cas("+",to_cascard t1, to_cascard t2)
      | to_cascard (Mul tr) = cas_list("*",map to_cascard tr)
      | to_cascard (Id s) = Var s
      | to_cascard (My(s,t)) = Mu(s,to_cascard t)
      | to_cascard (Parser.Lambda(s,t)) = Lambda(s,to_cascard t)

    val typ'    = to_cascard typ       (* Den Typ kaskardieren      *)
    val _       = print ("input= " ^ TypToString typ' ^ "\n")
    val tree    = enter typ'           (* den zyklischen Graphen in die Datenbank einfuegen    *)
in
    print ("db = " ^ TypToString(return tree) ^ "\n");
    print ("id = " ^ (Int.toString tree)  ^ "\n"); 
    SOME tree
end

fun exec s = 
let
    val tl = lex s                 (* zuerst den Lexer aufrufen *)
in
    if is1 "help" tl then
	(print( "mode standard  : Wechelt in den Standardmodus\n"^
	       "mode interface : Verwendet das vorgegebene Interface\n"^
	       "Syntax fr Eingaben: [name = ] type\n" ^
	       "                     type = unit | null | x | \n"^
	       "                            type + type | type -> type | \n"^
	       "                            type * type * .. * type |\n"^
	       "                            l x. type | u x. type\n"^
	       "verify rep     : Fhrt einen Test der Datenbankalgorithmen durch\n"^
	       "verify share   : Dito fr den Share Algorithmus\n"^
	       "speed test     : Geschwindigkeitsvergleich zwischen dem \n"^
	       "                 naiven Share und dem Partitioning Algorithmus \n"^
	       " q oder quit   : Beendet das Programm\n"^
	       "Im Interface Modus kann durch Eingabe einer Zahl die zugehrige\n"^
	       "Partition ausgegeben werden\n" );NONE)
    else if is2 "verify" "share" tl then (verify_share(); NONE)
    else if is2 "mode" "standard" tl then (print "neuer Modus: STANDARD\n"; mode := STANDARD; NONE)
    else if is2 "mode" "interface" tl then (print "neuer Modus: INTERFACE\n"; mode := INTERFACE; NONE)
    else if is2 "verify" "rep" tl then (verify_rep(); NONE)
    else if is2 "speed" "test" tl then (speed_test(); NONE) 
    else if is2 "speed" "save" tl then (speed_save ("naive_share.csv",1,1000,10000); NONE)
    else 
	let
	    val (tl',name) = parse_name tl (* Zuerst nachschauen, ob ein Name angegeben wurde...   *)
	    val typ = ty tl'               (* ... dann den Typ erstellen                           *)
	in
	    case typ of NONE => NONE       (* Wenn ein Fehler aufgetreten ist, nicht weitermachen. *)
	  |(SOME typ') => 
		case !mode of 
		    STANDARD => exec_standard typ' name
		  | INTERFACE => exec_iface typ'
	end
end

(* Das naechste Semikolon suchen *)
fun findSemi s n = if n = size s then ~1
              else if String.sub(s,n) = #";" then n
              else findSemi s (n+1)

(* Hauptfunktion *)
fun main s =
    let
	val _   = if s = "" then print "- " else print "  "
	val str = s ^ TextIO.inputLine TextIO.stdIn
    in
	if (str = "quit\n" orelse str = "q\n") then ()
	else case findSemi str 0 of
	    ~1 => main str 
	  |  n => let val _ = print "> "
		      val e = exec (String.substring(str,0,n)) 
		  in main ""  end
    end
handle NotClosed => (print "error: not closed\n"; main "")
handle InvalidType => (print "error: invalid type\n"; main "")
handle _ => (print "error: unknown error \n"; main "")

(* Start *)
val _ = 
(
 print ("\n\nTypendatenbank\n"^
        "--------------\n\n"^
	"Mit help; erhalten Sie Hilfe\n\n"^ 
        "Beispieleingaben:\n"^
        "nat = u t. unit + t;\n"^
        "unit + nat;\n"^
        "tree = u x. unit + nat*x;\n"^
        "unit + nat*tree\n\n");
 main "" 
)

end

