(**************************************************************
*
* share: Minimiert einen Baum
*
**************************************************************)

(* Lnge der Liste *)
fun ilength l = let val i = ref 0
                in (imap (fn n => i:= !i+1) l; !i) end
(* Ausgabe der textuellen Darstellung eines Blocks *)
fun drawBlocks (l:((''a Tree ref IList * int * ''a Tree ref option ref) ref IList)) = 
    (imap (fn n => print(Int.toString(ilength(#1(!(value(n)))))^" ")) l; print "\n")

(* jeder Block speichert die Knoten, die Anzahl deren Unterbume
   sowie den beim Zusammenbauen verwendeten Knoten *)
type ''a block = (''a Tree ref IList * int * ''a Tree ref option ref) ref

fun share t: (''a Tree ref) = 
    let
	(* Partitionierung der Knoten eines Baumes *)
	val blocks: (''a block) IList = empty()
	(* Liste der schon von 'initiateBlocks' bearbeiteten Knoten *)
	val done_nodes: (''a Tree ref) set ref = ref emptyset
        (* Findet den ersten Block, in dem ein Knoten mit dem geg. Label und #Nachfolger eingetragen ist*)
        val label2block: (''a * int, ''a block) map ref = ref emptymap
        (* Findet den Block, in dem ein Knoten eingetragen ist *)
	val node2block:  (''a Tree ref, ''a  block) map ref = ref emptymap

	(* Sortiert die Knoten des Baumes nach Labels und Zahl der Unterbume geordnet in 'blocks' ein*)
	fun initiateBlocks n =
	    if not (member(!done_nodes,n)) then
		( done_nodes := Set.insert(!done_nodes,n);
		 let 
		     val numtrees = numberOfSubtrees(n);
		     val labelInBlock = lookup(!label2block,(label(n),numtrees))
		 in  case labelInBlock of
		     NONE => let val new:''a Tree ref IList = empty()
				 val newblock = ref (new,numtrees,ref NONE)
			     in  
				 pushBack(new,n);
				 pushBack(blocks, newblock);
				 node2block := insertmap(!node2block,(n,newblock));
				 label2block := insertmap(!label2block,((label(n),numtrees),newblock))
			     end
		   | SOME b => (pushBack(#1(!b),n);
				node2block := insertmap(!node2block,(n,b)))
		 end;
		 case !n of
		     E => ()
		   | (N (_,l)) => (map initiateBlocks l; ())
	        )
	    else ()
 
        (* Verfeinert die Partitionierung, bis nur noch quivalente Knoten  *)
        (* in demselben Block gespeichert sind                              *)
	fun splitBlocks () =
	    let 
		val something_happened = ref true;

		(* Ordnet den Knoten in einem Block neue Blocks zu          *)
		(* Diese Funktion wird dann auf die ganze Partition gemappt *)
		fun f blockNode =
		    let
			val block = value(blockNode)
			val numtrees = #2(!block)
			val nextblocks = createhashtable()
			val firstNode = ref true
			val new_node2block = ref nil

			fun update_node2block nil = ()
			  | update_node2block ((n,b)::xr) = 
			    (node2block := insertmap(!node2block,(n,b));
			     update_node2block xr)

			fun g node = 
			    let 
				fun search x = 
				    case lookup(!node2block,x) of
					NONE     => raise Empty
				      | (SOME y) => y

				val nextblockslist:''a block list =
				    case !(value(node)) of
					(N(_,xs)) => map search xs
				      | _         => nil
			    in
				case lookuphash nextblockslist nextblocks of
				    NONE => 
					(case !firstNode of
					     false => let 
							  val newblock:(''a block) = 
							      ref (empty(),
								   numberOfSubtrees((value(node))),
								   ref NONE)
						      in
							  pushBack(#1(!newblock),value(node));
							  new_node2block := (value(node), newblock)::(!new_node2block);
							  inserthash (nextblockslist,newblock) nextblocks;
							  pushBack(blocks,newblock);
							  erase(#1(!block),node);
							  something_happened := true
						      end
					   | true =>  (inserthash (nextblockslist,block) nextblocks;
						       firstNode := false ) )
				  | (SOME b) => (pushFront(#1(!b),value(node));
						 new_node2block := (value(node), b)::(!new_node2block);
						 erase(#1(!block),node); ())
			    end
		    in
			imap g (#1(!block));
			update_node2block (!new_node2block)
		    end
	    in
		while !something_happened do
		    (something_happened := false;
		     (*drawBlocks blocks;*)
		     imap f blocks )
		(*drawBlocks blocks*)
	    end;


        (* Baut mit Hilfe der Blocks aus dem alten Baum einen neuen, reduzierten auf. *)
        (* Dazu wird aus jedem Block ein Knoten als Reprsentant gewhlt und alle     *)
        (* quivalenten Knoten durch diesen ersetzt.                                  *)
	fun reduceTree (t:''a Tree ref) =
	    (* In welchem Block steht der Knoten? *)
	    case lookup(!node2block,t) of
		NONE     => raise Empty
	      | (SOME b) => 
		    (* Haben wir schon einen Knoten aus diesem Block benutzt? *)
		    (case !(#3(!b)) of
			 NONE      => (#3(!b) := SOME t;
				       (case !t of 
					    (N(l,xs)) => t := N(l,map reduceTree xs)
					  | _         => ());
					t)
		       | (SOME tr) => tr ); 
    in 
	initiateBlocks(t);
	splitBlocks();
	reduceTree(t)
    end
